/*  gngeo a neogeo emulator
 *  Copyright (C) 2001 Peponas Mathieu
 * 
 *  This program is free software; you can redistribute it and/or modify  
 *  it under the terms of the GNU General Public License as published by   
 *  the Free Software Foundation; either version 2 of the License, or    
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
#include "getopt.h"
#endif


#include "SDL.h"
#include "conf.h"
#include "fileio.h"

/*
static CONF_ITEM *cf_hash[128];
static cf_hash_size[128];
*/

static struct {
    CONF_ITEM **conf;
    int size,nb_item;
}cf_hash[128];

static int default_key1[] = { 119, 120, 113, 115, 38, 34, 273, 274, 276, 275 };
static int default_key2[] = { 108, 109, 111, 112, 233, 39, 264, 261, 260, 262 };
static int default_joy2[] = { 1, 0, 3, 2, 7, 6, 0, 1 };
static int default_joy1[] = { 2, 3, 0, 1, 5, 4, 0, 1 };


static CONF_ITEM * create_conf_item(char *name,char *help,char short_opt)
{
    int a;
    static int val=0x100;
    CONF_ITEM *t=(CONF_ITEM*)calloc(1,sizeof(CONF_ITEM));
    
    a=tolower ((int)name[0]);

    t->name=name;
    t->help=help;
    if (short_opt==0) {
	val++;
	t->short_opt=val;
    } else
	t->short_opt=short_opt;

    
    if (cf_hash[a].size<=cf_hash[a].nb_item) {
	cf_hash[a].size+=10;
	cf_hash[a].conf=(CONF_ITEM**)realloc(cf_hash[a].conf,cf_hash[a].size*sizeof(CONF_ITEM*));
    }

    cf_hash[a].conf[cf_hash[a].nb_item]=t;
    cf_hash[a].nb_item++;
    return t;
}

void cf_create_bool_item(char *name,char *help,char short_opt,SDL_bool def)
{
    CONF_ITEM *t=create_conf_item(name,help,short_opt);
    t->type=CFT_BOOLEAN;
    t->data.dt_bool.bool=def;
    t->data.dt_bool.default_bool=def;
}

void cf_create_string_item(char *name,char *help,char short_opt,char *def)
{
    CONF_ITEM *t=create_conf_item(name,help,short_opt);
    t->type=CFT_STRING;
    strcpy(t->data.dt_str.str,def);
    t->data.dt_str.default_str=def;
}

void cf_create_int_item(char *name,char *help,char short_opt,int def)
{
    CONF_ITEM *t=create_conf_item(name,help,short_opt);
    t->type=CFT_INT;
    t->data.dt_int.val=def;
    t->data.dt_int.default_val=def;
}

void cf_create_array_item(char *name,char *help,char short_opt,int size,int *def)
{
    CONF_ITEM *t=create_conf_item(name,help,short_opt);
    t->type=CFT_ARRAY;
    t->data.dt_array.size=size;
    t->data.dt_array.array=(int*)calloc(1,size*sizeof(int));
    memcpy(t->data.dt_array.array,def,size*sizeof(int));
    t->data.dt_array.default_array=def;
}

CONF_ITEM* cf_get_item_by_name(char *name){
    int i;
    int a=tolower ((int)name[0]);

    for(i=0;i<cf_hash[a].nb_item;i++) {
	if (strcasecmp(cf_hash[a].conf[i]->name,name)==0)
	    return cf_hash[a].conf[i];
    }
    return NULL;
}

CONF_ITEM* cf_get_item_by_val(int val){
    int i,j;

    for(i=0;i<128;i++) {
	for(j=0;j<cf_hash[i].nb_item;j++) {
	    if (cf_hash[i].conf[j]->short_opt==val)
		return cf_hash[i].conf[j];
	}
    }
    return NULL;
}

void cf_init(void)
{
    CONF_ITEM *t;

    cf_create_bool_item("forcepc","Force the PC to a correct value at startup",0,SDL_FALSE);
    cf_create_bool_item("fullscreen","Start gngeo in fullscreen",'f',SDL_FALSE);
    cf_create_bool_item("interpolation","Merge the last frame and the current",'I',SDL_FALSE);
    cf_create_bool_item("raster","Enable the raster interrupt",'r',SDL_FALSE);
    cf_create_bool_item("sound","Enable sound",0,SDL_TRUE);
    cf_create_bool_item("showfps","Show FPS at startup",0,SDL_FALSE);
    cf_create_bool_item("autoframeskip","Enable auto frameskip",0,SDL_TRUE);
    cf_create_bool_item("sleepidle","Sleep when idle",0,SDL_FALSE);
    cf_create_bool_item("joystick","Enable joystick support",0,SDL_TRUE);
    cf_create_bool_item("invertjoy","Invert joystick order",0,SDL_FALSE);
    cf_create_bool_item("debug","Start with inline debuger",'D',SDL_FALSE);
    cf_create_bool_item("hwsurface","Use hardware surface for the screen",'H',SDL_FALSE);
    cf_create_bool_item("convtile","Convert tile in internal format at loading",'c',SDL_TRUE);
    cf_create_bool_item("pal","Use PAL timing (buggy)",'P',SDL_FALSE);

    cf_create_string_item("country","Set the contry to japan, usa or europe",0,"japan");
    cf_create_string_item("system","Set the system to home or arcade",0,"arcade");
    cf_create_string_item("rompath","Use STRING as rom path",'i',DATA_DIRECTORY);
    cf_create_string_item("romrc","Use STRING as romrc file",'d',DATA_DIRECTORY"/romrc");
    cf_create_string_item("libglpath","Use STRING as libGL",0,"/usr/lib/libGL.so");
    cf_create_string_item("effect","Use the specified effect (help for a list)",'e',"none");
    cf_create_string_item("blitter","Use the specified blitter (help for a list)",'b',"soft");
    cf_create_string_item("transpack","Use the specified transparency pack",'t',"none");
    
    cf_create_array_item("p1key","Player1 Keyboard configuration",0,10,default_key1);
    cf_create_array_item("p2key","Player2 Keyboard configuration",0,10,default_key2);
    cf_create_array_item("p1joy","Player1 Joystick configuration",0,8,default_joy1);
    cf_create_array_item("p2joy","Player2 Joystick configuration",0,8,default_joy2);

    cf_create_int_item("scale","Scale the resolution by N",0,1);
    cf_create_int_item("samplerate","Set the sample rate to N",0,22050);

}

SDL_bool discard_line(char *buf)
{
    if (buf[0] == '#')
	return SDL_TRUE;
    if (buf[0] == '\n')
	return SDL_TRUE;
    return SDL_FALSE;
}

static void read_array(int *tab, char *val, int size) {
    int i = 0;
    char *v;

    v = strtok(val, ",");

    while (v != NULL && i < size) {
	tab[i] = atoi(v);
	v = strtok(NULL, ",");
	i++;
    }
}

SDL_bool cf_open_file(char *filename)
{
    /* if filename==NULL, we use the default one: $HOME/.gngeo/gngeorc */
    char *conf_file=filename;
    FILE *f;
    int i=0;
    char buf[512];
    char name[32];
    char val[255];
    CONF_ITEM *cf;

    if (!conf_file) {
	int len = strlen("gngeorc") + strlen(getenv("HOME")) + strlen("/.gngeo/") +	1;
	conf_file = (char *) alloca(len*sizeof(char));
	sprintf(conf_file, "%s/.gngeo/gngeorc", getenv("HOME"));
    }
    if ((f = fopen(conf_file, "rb")) == 0) {
	printf("Unable to open %s\n",conf_file);
	return SDL_FALSE;
    }

    while (!feof(f)) {
	i = 0;
	fgets(buf, 510, f);
	if (discard_line(buf))
	    continue;
	sscanf(buf, "%s %s\n", name, val);
	cf=cf_get_item_by_name(name);
	if (cf) {
	    /*printf("Option %s\n",cf->name);*/
	    switch(cf->type) {
	    case CFT_INT:
		CF_VAL(cf)=atoi(val);
		break;
	    case CFT_BOOLEAN:
		CF_BOOL(cf)=(strcasecmp(val,"true")==0?SDL_TRUE:SDL_FALSE);
		break;
	    case CFT_STRING:
		strncpy(CF_STR(cf),val,254);
		break;
	    case CFT_ARRAY:
		read_array(CF_ARRAY(cf),val,CF_ARRAY_SIZE(cf));
		break;
	    }
	} else {
	    /*printf("Unknow option %s\n",name);*/
	    /* unknow option...*/
	}
    }
    return SDL_TRUE;
}

void cf_print_help(void) {
    int i,j;
    CONF_ITEM *cf;
    printf("Usage: gngeo [OPTION]... ROMSET\n"
	   "Emulate the NeoGeo rom designed by ROMSET\n\n");
    printf("  -h, --help                 Print this help and exit\n"
	   "  -l, --listgame             List all the available game\n");
    for (i=0;i<128;i++) {
	for(j=0;j< cf_hash[i].nb_item;j++) {
	    cf=cf_hash[i].conf[j];
	    if (cf->short_opt<128 && cf->short_opt>=32)
		printf("  -%c, --",cf->short_opt);
	    else
		printf("      --");
	    switch (cf->type) {
	    case CFT_ARRAY:
	    {
		char buf[22];
		snprintf(buf,21,"%s=ARRAY",cf->name);
		printf("%-20s %s\n",buf,cf->help);
	    }
	    break;
	    case CFT_STRING:
	    {
		char buf[22];
		snprintf(buf,21,"%s=STRING",cf->name);
		printf("%-20s %s\n",buf,cf->help);
	    }
	    break;
	    case CFT_INT:
	    {
		char buf[22];
		snprintf(buf,21,"%s=N",cf->name);
		printf("%-20s %s\n",buf,cf->help);
	    }
		break;
	    case CFT_BOOLEAN:
		printf("%-20s %s\n",cf->name,cf->help);
		break;
	    }
	}
    }
    printf("\nAll boolean options can be disabled with --no-OPTION\n"
	   "(Ex: --no-sound turn sound off)\n\n");
}

static struct option *longopt;
static void add_long_opt_item(char *name,int has_arg,int *flag,int val) {
    static int opt_size=0;
    static int opt=0;

    if (opt>=opt_size) {
	opt_size+=10;
	longopt=realloc(longopt,(opt_size+1)*sizeof(struct option));
    }
    
    longopt[opt].name=name;
    longopt[opt].has_arg=has_arg;
    longopt[opt].flag=flag;
    longopt[opt].val=val;

    opt++;
}

int option_index = 0;
char* cf_parse_cmd_line(int argc, char *argv[]){

    char shortopt[255];
    //int opt_size=10,opt=-1;
    int i,j,c;
    CONF_ITEM *cf;
    char *sbuf;
    int buflen;

    memset(shortopt,0,255);
    
    for (i=0;i<128;i++) {
	for(j=0;j< cf_hash[i].nb_item;j++) {
	    cf=cf_hash[i].conf[j];
	    
	    if (cf->short_opt<=128) {
		char b[2];
		sprintf(b,"%c",cf->short_opt);
		strcat(shortopt,b);
	    }

	    switch(cf->type) {
	    case CFT_ARRAY:
	    case CFT_STRING:
	    case CFT_INT:
		if (cf->short_opt<=128) strcat(shortopt,":");
		add_long_opt_item(cf->name,1,NULL,cf->short_opt);
	    break;
	    case CFT_BOOLEAN:
		add_long_opt_item(cf->name,0,&CF_BOOL(cf),1);
		/* create the --no-option */
		buflen=strlen("no-")+strlen(cf->name);
		sbuf=malloc(buflen+2);
		snprintf(sbuf,buflen+1,"no-%s",cf->name);
		add_long_opt_item(sbuf,0,&CF_BOOL(cf),0);
		break;
	    }
	}/* for j*/
    }/* for i*/
    
    /* not really option, but have to be available on the cmd line*/
    add_long_opt_item("help",0,NULL,'h');
    add_long_opt_item("listgame",0,NULL,'l');
    strcat(shortopt,"hl");

    /* end the longopt array*/
    add_long_opt_item(0,0,0,0);
    //printf("shortopt:%s\n",shortopt);
    
    while((c=getopt_long(argc,argv,shortopt,longopt, &option_index))!=EOF) {
	switch (c) {
	case 'h': 
	    cf_print_help();
	    exit(0);
	    break;
	case 'l':
	    list_game();
	    exit(0);
	    break;
	}
	if (c!=0) {
	    cf=cf_get_item_by_val(c);
	    if (cf) {
		switch(cf->type) {
		case CFT_INT:
		    CF_VAL(cf)=atoi(optarg);
		    break;
		case CFT_BOOLEAN:
		    CF_BOOL(cf)=1;
		    break;
		case CFT_STRING:
		    strcpy(CF_STR(cf),optarg);
		    break;
		case CFT_ARRAY:
		    read_array(CF_ARRAY(cf),optarg,CF_ARRAY_SIZE(cf));
		    break;
		}
	    }
	}
    }
    if (optind >= argc)
	return NULL;
    return strdup(argv[optind]);
}
